; Copyright  2004 Ben J. Jackson <ben@ben.com>
; May be freely used, modified and distributed,
; provided this notice remains intact.  This software
; is provided "as is" without express or implied warranty.

; EXAMPLE:
;
; include "18fdiv.inc"
; cblock 0x80 ; each arg must be contiguous,
;             ; but the args don't have to be adjacent
;  ARGA:4 ; dividend, MSB in ARGA, LSB in ARGA+n
;  ARGB:2 ; divisor
;  REM:2 ; remainder (output, same size as divisor)
;  LOOP
; endc
; mkdiv foo, ARGA, ARGB, REM, LOOP, d'32', d'16', UNSIGNED
;
; Makes a function foo32x16u.  If you don't need all 8*n bits
; of the divisor, you can reduce it by one to get a slightly
; faster function:
;
; mkdiv foo, ARGA, ARGB, REM, LOOP, d'32', d'15', UNSIGNED
;
; (No other shortening of the bits helps or is allowed.)
;
; If you need a signed divider, change the last argument:
;
; mkdiv foo, ARGA, ARGB, REM, LOOP, d'24', d'7', SIGNED
;
; This produces foo24x7s.  Notice you can re-use the blocks
; of memory as long as they are at least long enough to fit.
; With signed divide it makes a lot of sense to drop a bit on
; the divisor, since only -2^n-1 doesn't fit once the divider
; takes the absolute value.
;
; BUGS:  The remainder (in rem) will have the sign of the
; quotient, and have the magnitude of the unsigned division.
; However, it will not satisfy (N/M)*M+(N%M) == N for some
; combinations of signs (for example 3/-2=-1, remainder -1,
; however -2*-1 -1=1, not 3, the remainder should be +1).
; It's not that hard to fix but I don't care about the remainder.
; It would take several instructions, a few cycles, and a bit
; to store some state.  Who does negative modulo anyway?
;
SIGNED equ 1
UNSIGNED equ 0

divneg_ macro aa, awords
	; note WREG must be clear if awords > 1
	local i = 0
	while i < awords - 1
		comf aa+#v(i)
		local i = i + 1
	endw
	negf aa+#v(i)
	while i > 0
		local i = i - 1
		addwfc aa+#v(i)
	endw
	endm

mkdiv macro name, aa, bb, rem, loop, abits, bbits, signed
	if signed
name#v(abits)x#v(bbits)s:
	else
name#v(abits)x#v(bbits)u:
	endif
	local awords = (abits + 7) / 8
	local bwords = (bbits + 7) / 8
	if ((abits > d'127' && (bbits == bwords * 8)) || abits > d'256')
		error "18fdiv: dividend too long"
		; If you shorten the divisor by a bit you get a wider dividend
		; because LOOP.7 is freed up
	endif
	if (abits != awords * 8)
		error "18fdiv has no optimizations for odd bit length dividends."
		; This could be done by clearing carry and pre-rotating by
		; (awords * 8 - abits) and then doing the normal divide.
	endif
	if (bbits < bwords * 8 - 1)
		messg "18fdiv has no optimizations for divisors any shorter than 8n-1."
		; The loop is simplified by not having the MSB of the divisor count,
		; but nothing any shorter than that triggers any optimizations.
	endif
	movlw abits
	movwf loop
	if signed
		; signed div will find the sign of the result, stash
		; it away, then take the absolute values of the terms,
		; do the divide, and restore the sign to the result and
		; remainder.
		movf bb, W
		xorwf aa, W
		movwf rem ; sign in bit 7
		if (awords > 1 || bwords > 1)
			clrf WREG
		endif
		btfsc aa, 7
		if (awords > 1)
			rcall name#v(abits)x#v(bbits)nega ; re-use
		else
			divneg_ aa, awords
		endif
		if (bwords > 1)
			btfss bb, 7
			bra name#v(abits)x#v(bbits)pd
		else
			btfsc bb, 7 ; it will only be 1 instr
		endif
		divneg_ bb, bwords
name#v(abits)x#v(bbits)pd:
		rlcf rem, W ; sign in C
	endif
	local i = bwords
	while i > 0
		local i = i - 1
		clrf rem+#v(i)
	endw
name#v(abits)x#v(bbits)l#v(signed):
	local i = awords
	while i > 0
		local i = i - 1
		rlcf aa+#v(i), F
	endw
	local i = bwords
	while i > 0
		local i = i - 1
		rlcf rem+#v(i), F
	endw

	; if we're using all 2^n bits, use loop.7 to save extra bit	
	if (bbits == bwords * 8)
		btfsc STATUS, C
		bsf LOOP, 7
	endif
	
	local i = bwords-1
	movf bb+#v(i), W
	subwf rem+#v(i), F
	while i > 0
		local i = i - 1
		movf bb+#v(i), W
		subwfb rem+#v(i), F
	endw
	; if no borrow, shortcut everything and get that C into answer
	bc name#v(abits)x#v(bbits)yy#v(signed)
	if (bbits == bwords * 8)
		; if a saved bit, try the rescue
		rlcf loop, W ; must get it in C for branch
		bc name#v(abits)x#v(bbits)yy#v(signed)
	endif
	; restore
	local i = bwords-1
	if i > 0 || (bbits == bwords * 8); else it's still in w
		movf bb+#v(i), W
	endif
	addwf rem+#v(i), F
	while i > 0
		local i = i - 1
		movf bb+#v(i), W
		addwfc rem+#v(i), F
	endw
	bcf STATUS, C
name#v(abits)x#v(bbits)yy#v(signed)
	if (bbits == bwords * 8)
		bcf LOOP, 7
	endif
	decfsz loop, F
	bra name#v(abits)x#v(bbits)l#v(signed)

	; rotate in the final result bit
	local i = awords
	while i > 0
		local i = i - 1
		rlcf aa+#v(i), F
	endw
	if signed
		; we stuffed the sign into the result and just rolled it off
		btfss STATUS, C
		return
		if (awords > 1 || bwords > 1)
			clrf WREG
		endif
		divneg_ rem, bwords
name#v(abits)x#v(bbits)nega
		divneg_ aa, awords
	endif
	return
	endm
